﻿//////////////////////////////////////////////
// LinearPager.h
//
//////////////////////////////////////////////

/// Defines / Macros -------------------------

#pragma once

/// Includes ---------------------------------

// nkMemory
#include "../../Log/LogManager.h"

#include "../MemorySpaces/DefaultMemorySpace.h"

// Standards
#include <deque>
#include <string>

/// Class ------------------------------------

namespace nkMemory
{
	// T = return type des allocations
	// U = memory space à utiliser
	template <typename T = char*, typename U = DefaultMemorySpace>
	class LinearPager final
	{
		private :
			
			// Attributs
			// Les pages en mémoire
			std::deque<U> _pages ;
			std::deque<U*> _availablePages ;

			// Taille des spaces en page
			size_t _pageSize ;
			// Offset actuel en mémoire
			size_t _currentOffset ;

		private :

			// Functions
			// Allocation de page
			void _allocatePage ()
			{
				_pages.emplace_back(_pageSize) ;
				_availablePages.emplace_back(&_pages.back()) ;
			}

		public :

			// Constructeur
			LinearPager (size_t pageSize)
			:	_pages (),
				_pageSize (pageSize),
				_currentOffset (0)
			{
				// Première page pour simplifier la gestion plus tard
				_allocatePage() ;
			}

			// Destructeur
			~LinearPager () = default ;

		public :

			// Getters
			size_t getPageSize () const
			{
				return _pageSize ;
			}

			size_t getCurrentOffset () const
			{
				return _currentOffset ;
			}

			size_t getAllocatedPageCount () const
			{
				return _pages.size() ;
			}

			size_t getAvailablePageCount () const
			{
				return _availablePages.size() ;
			}

		public :

			// Allocations
			// Obtention zone mémoire
			T allocate (size_t size, size_t alignment = 1)
			{
				// Check de taille
				if (size > _pageSize || alignment > _pageSize)
				{
					std::string message ;
					message += "Attempted to allocate a chunk of memory of size " ;
					message += std::to_string(size) ;
					message += " and alignment of " ;
					message += std::to_string(alignment) ;
					message += " on a Pager having pages of size " ;
					message += std::to_string(_pageSize) ;
					message += ". Request cannot be handled, returning nullptr." ;

					LogManager::getInstance()->log(message, "LinearPager") ;

					return T() ;
				}

				// On calcule l'adresse à prendre
				size_t toAlignOffset = _currentOffset % alignment ;
				size_t alignedOffset = toAlignOffset ? _currentOffset + (alignment - toAlignOffset) : _currentOffset ;

				// Check si on a la place
				size_t remainingSpace = getCurrentPageFreeMemory() ;

				if (remainingSpace < alignedOffset + size)
				{
					// Il nous faut une nouvelle page
					_availablePages.pop_front() ;

					if (_availablePages.empty())
						_allocatePage() ;

					// Reset des offsets
					_currentOffset = 0 ;
					alignedOffset = 0 ;
				}

				// Et on demande l'allocation
				T allocation = _availablePages.front()->getOffsetPtr(alignedOffset) ;
				_currentOffset = alignedOffset + size ;

				return allocation ;
			}

			// Reset de mémoire (état pages, ne libère aucune mémoire)
			void reset ()
			{
				// On reset notre état pour réutiliser les pages disponibles
				_availablePages.clear() ;

				for (U& entry : _pages)
					_availablePages.emplace_back(&entry) ;

				_currentOffset = 0 ;
			}

		public :

			// Utilitaires
			size_t getCurrentPageFreeMemory () const
			{
				return _pageSize - _currentOffset ;
			}
	} ;
}